/*************************************************************************/
/* fortran_and_c.c                                                       */
/*                                                                       */
/* fortran_and_c.a                                                 @ti@  */
/*                                                                       */
/* Author:                                                               */
/*   Arve Kylling                                                  @au@  */
/*   NILU,                                                         @ad@  */
/*   9005 Tromso,                                                  @ad@  */
/*   Norway                                                        @ad@  */
/*                                                                       */
/* ----------------------------------------------------------------------*/
/* Copyright (C) 1998 Arve Kylling                                       */
/*                                                                       */
/* This program is free software; you can redistribute it and/or modify  */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation; either version 1, or (at your option)   */
/* any later version.                                                    */
/*                                                                       */
/* This program is distributed in the hope that it will be useful,       */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the          */
/* GNU General Public License for more details.                          */
/*                                                                       */
/* To obtain a copy of the GNU General Public License write to the       */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,    */
/* USA.                                                                  */
/*-----------------------------------------------------------------------*/
/*************************************************************************/


/* @32c@ */
/* @c32@ */
/****************************************************************************/
/* In order to use the functions provided by the fortran_and_c    @32_10c@  */
/* library, #include <fortran_and_c.h> in your source code and link         */
/* with fortran_and_c.a.                                                    */
/*                                                                          */
/* @strong{Example:}                                                        */
/* Example for a source file:                                               */
/* @example                                                                 */
/*                                                                          */
/*   ...                                                                    */
/*   #include <fortran_and_c.h>                                             */
/*   ...                                                                    */
/*                                                                          */
/* @end example                                                             */
/*                                                                          */
/* Linking of the executable, using the GNU compiler gcc:                   */
/* @example                                                                 */
/*                                                                          */
/*   gcc -o test test.c -lfortran_and_c                                     */
/*                                                                          */
/* @end example                                                   @c32_10@  */
/****************************************************************************/


/****************************************************************************/
/* The fortran_and_c library provides functions converting        @32_20c@  */
/* from multidimensional C arrays to one-dimensional fortran arrays         */
/* that can be input to fortran routines, and functions for converting      */
/* one-dimensional fortran compatible arrays to multidimensional C arrays.  */
/*                                                                          */
/* These functions are useful when calling fortran functions and            */
/* subroutines from C.                                                      */
/*                                                                 @c32_20@ */
/****************************************************************************/

/***********************************************************************************/
/* Function: c2fortran_2D_float_ary                                       @32_30i@ */
/* Description:                                                                    */
/* Convert 2D C float array to 1D column float array and map it such that it       */
/* can be passed to a fortran routine.                                             */
/*                                                                                 */
/* Parameters:                                                                     */
/*  int dim1:          first dimension of C array                                  */
/*  int dim2:          second dimension of C array                                 */
/*  float **ary_2D:    pointer to the 2D C array                                   */
/*                                                                                 */
/* Return value:                                                                   */
/*  Pointer to the 1D fortran compatible array. Memory allocation is done          */
/*  automatically and can be freed with a simple free().                           */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author: Arve Kylling                                                            */
/*                                                                        @i32_30@ */
/***********************************************************************************/

#include "fortran_and_c.h"

float* c2fortran_2D_float_ary(int dim1, int dim2, float** ary_2D ){
  
  int i1, i2, j;
  float *tmp = (float *) calloc (dim1*dim2, sizeof(float));
  j = 0;
  for (i1=0;i1<dim1;i1++) {
    for (i2=0;i2<dim2;i2++) {
      tmp[j] = ary_2D[i1][i2];
      j++;
    }
  }
  return tmp;
}

/***********************************************************************************/
/* Function: fortran2c_2D_float_ary                                       @32_30i@ */
/* Description:                                                                    */
/* Convert 1D fortran compatible float array to 2D C float array. Space for the    */
/* returned ary is automatically allocated.                                        */
/*                                                                                 */
/* Parameters:                                                                     */
/*  int dim1:          first dimension of C array                                  */
/*  int dim2:          second dimension of C array                                 */
/*  float *ary_1D:     pointer to the 1D fortran compatible array                  */
/*                                                                                 */
/* Return value:                                                                   */
/*  Pointer to the 2D C array. Memory allocation is done automatically.            */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author: Arve Kylling                                                            */
/*                                                                        @i32_30@ */
/***********************************************************************************/
float** fortran2c_2D_float_ary(int dim1, int dim2, float* ary_1D ){
  
  int i1, i2, j, status=0;
  float **tmp;
  status = ASCII_calloc_float (&tmp, dim1, dim2);
  j = 0;
  for (i1=0;i1<dim1;i1++) {
    for (i2=0;i2<dim2;i2++) {
       tmp[i1][i2] = ary_1D[j++];
    }
  }
  return tmp;
}

/***********************************************************************************/
/* Function: fortran2c_3D_float_ary                                       @32_30i@ */
/* Description:                                                                    */
/* Convert 1D fortran compatible float array to 3D C float array. Space for the    */
/* returned ary is automatically allocated.                                        */
/*                                                                                 */
/* Parameters:                                                                     */
/*  int dim1:          first dimension of C array                                  */
/*  int dim2:          second dimension of C array                                 */
/*  float *ary_1D:     pointer to the 1D fortran compatible array                  */
/*                                                                                 */
/* Return value:                                                                   */
/*  Pointer to the 3D C array. Memory allocation is done automatically.            */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author: Arve Kylling                                                            */
/*                                                                        @i32_30@ */
/***********************************************************************************/
float*** fortran2c_3D_float_ary(int dim1, int dim2, int dim3, float* ary_1D ){
  
  int i1, i2, i3, j, status=0;
  float ***tmp;
  status = ASCII_calloc_float_3D (&tmp, dim1, dim2, dim3);
  j = 0;
  for (i1=0;i1<dim1;i1++) {
    for (i2=0;i2<dim2;i2++) {
      for (i3=0;i3<dim3;i3++) {
	tmp[i1][i2][i3] = ary_1D[j++];
      }
    }
  }
  return tmp;
}

/***********************************************************************************/
/* Function: fortran2c_4D_float_ary                                       @32_30i@ */
/* Description:                                                                    */
/* Convert 1D fortran compatible float array to 4D C float array. Space for the    */
/* returned ary is automatically allocated.                                        */
/*                                                                                 */
/* Parameters:                                                                     */
/*  int dim1:          first dimension of C array                                  */
/*  int dim2:          second dimension of C array                                 */
/*  float *ary_1D:     pointer to the 1D fortran compatible array                  */
/*                                                                                 */
/* Return value:                                                                   */
/*  Pointer to the 4D C array. Memory allocation is done automatically.            */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author: Arve Kylling                                                            */
/*                                                                        @i32_30@ */
/***********************************************************************************/
float**** fortran2c_4D_float_ary(int dim1, int dim2, int dim3, int dim4, float* ary_1D ){
  
  /* Convert 1D fortran compatible array to 4D C array. Space for the 
     returned ary is allocated. */

  int i1, i2, i3, i4, j, status=0;
  float ****tmp;
  status = ASCII_calloc_float_4D (&tmp, dim1, dim2, dim3, dim4);
  j = 0;
  for (i1=0;i1<dim1;i1++) {
    for (i2=0;i2<dim2;i2++) {
      for (i3=0;i3<dim3;i3++) {
	for (i4=0;i4<dim4;i4++) {
	  tmp[i1][i2][i3][i4] = ary_1D[j++];
	}
      }
    }
  }
  return tmp;
}

